#include <PSX/GPU/gpu.hpp>
#include <PSX/GPU/gpu_primitives.hpp>
#include <FontWriter/font_writer.hpp>
#include <stdio.h>

#include <PSX/Auxiliary/mem_dump.hpp>

namespace JabyEngine {
    struct DoubleBuffer {
        GPU::TexPage::Linked tex_page;
        FontBufferInfo       prim_buffer;
        FontPrimitive*       cur_primitive;
        
        static constexpr DoubleBuffer empty() {
            return DoubleBuffer{.tex_page = {0}, .prim_buffer = FontBufferInfo::empty(), .cur_primitive = nullptr};
        }

        void setup(const FontBufferInfo& buffer_info, const SimpleTIM& vram_dst, const GPU::SizeI16& font_size) {
            this->tex_page      = Make::TexPage(vram_dst.get_texture_position(), GPU::TexturePageColor::$4bit).linked();
            this->prim_buffer   = buffer_info;
            this->cur_primitive = buffer_info.double_buffer[GPU::Display::current_id];

            for(size_t buffer_id = 0; buffer_id < 2; buffer_id++) {
                for(size_t buffer_element_id = 0; buffer_element_id < buffer_info.single_buffer_length; buffer_element_id++) {
                    this->prim_buffer.double_buffer[buffer_id][buffer_element_id] = Make::SPRT(
                        Make::AreaI16(Make::PositionI16(0, 0), font_size),
                        Make::OffsetPageWithClut(Make::PageOffset(0, 0), Make::PageClut(vram_dst.get_clut_position()))
                    ).linked();
                }
            }
        }

        void write(GPU::PositionI16 pos, const char* str) {
            const auto* primitive_end = &this->prim_buffer.double_buffer[GPU::Display::current_id][this->prim_buffer.single_buffer_length];

            while(this->cur_primitive < primitive_end && *str != '\0') {
                auto& primitive = *this->cur_primitive;

                this->cur_primitive++;
                str++;
                primitive->position = pos;
                primitive.concat(*this->cur_primitive);

                pos.move(primitive->size.width, 0);
            }
        }

        void render() {
            const auto update_id = GPU::Display::current_id;
            const auto render_id = update_id ^ 1;

            this->cur_primitive -= 1;
            if(this->cur_primitive >= this->prim_buffer.double_buffer[render_id]) {
                this->cur_primitive->terminate();
                this->tex_page.concat(*this->prim_buffer.double_buffer[render_id]);

                GPU::render(this->tex_page);
            }

            this->cur_primitive = this->prim_buffer.double_buffer[update_id];
        }
    };

    static auto double_buffer = DoubleBuffer::empty();

    void FontWriter::Cursor :: write(const char* str) {
        double_buffer.write(this->pos, str);
    }

    void FontWriter :: setup(const FontBufferInfo& buffer_info, const SimpleTIM& vram_dst, const FontInfo& font_info) {
        double_buffer.setup(buffer_info, vram_dst, Make::SizeI16(font_info.FontSize.width, font_info.FontSize.height));
    }

    FontWriter::Cursor FontWriter :: start(const GPU::PositionI16& pos) {
        return Cursor{.pos = pos};
    }

    void FontWriter :: render() {
        double_buffer.render();
    }
}